import os.path
from arcana.exception import ArcanaError
import re


PATH_SUFFIX = '_path'
FIELD_SUFFIX = '_field'

package_dir = os.path.join(os.path.dirname(__file__), '..')


def is_regex(s):
    "Checks to see if string contains special characters"
    return bool(re.match(r'^\w+$', s))


def nth(i):
    "Returns 1st, 2nd, 3rd, 4th, etc for a given number"
    if i == 1:
        s = '1st'
    elif i == 2:
        s = '2nd'
    elif i == 3:
        s = '3rd'
    else:
        s = '{}th'.format(i)
    return s


def dir_modtime(dpath):
    """
    Returns the latest modification time of all files/subdirectories in a
    directory
    """
    return max(os.path.getmtime(d) for d, _, _ in os.walk(dpath))


def get_fsl_reference_path():
    return os.path.join(os.environ['FSLDIR'], 'data', 'standard')


def get_atlas_path(name, dataset='brain', resolution='1mm'):
    """
    Returns the path to the atlas (or atlas mask) in the arcana repository

    Parameters
    ----------
    name : str
        Name of the Atlas, can be one of ('mni_nl6')
    atlas_type : str
        Whether to return the brain mask or the full atlas, can be one of
        'image', 'mask'
    """
    if name == 'MNI152':
        # MNI ICBM 152 non-linear 6th Generation Symmetric Average Brain
        # Stereotaxic Registration Model (http://nist.mni.mcgill.ca/?p=858)
        if resolution not in ['0.5mm', '1mm', '2mm']:
            raise ArcanaError(
                "Invalid resolution for MNI152, '{}', can be one of '0.5mm', "
                "'1mm' or '2mm'".format(resolution))
        if dataset == 'image':
            path = os.path.join(get_fsl_reference_path(),
                                'MNI152_T1_{}.nii.gz'.format(resolution))
        elif dataset == 'mask':
            path = os.path.join(get_fsl_reference_path(),
                                'MNI152_T1_{}_brain_mask.nii.gz'
                                .format(resolution))
        elif dataset == 'mask_dilated':
            if resolution != '2mm':
                raise ArcanaError(
                    "Dilated MNI masks are not available for {} resolution "
                    .format(resolution))
            path = os.path.join(get_fsl_reference_path(),
                                'MNI152_T1_{}_brain_mask_dil.nii.gz'
                                .format(resolution))
        elif dataset == 'brain':
            path = os.path.join(get_fsl_reference_path(),
                                'MNI152_T1_{}_brain.nii.gz'
                                .format(resolution))
        else:
            raise ArcanaError("Unrecognised dataset '{}'"
                                  .format(dataset))
    else:
        raise ArcanaError("Unrecognised atlas name '{}'"
                              .format(name))
    return os.path.abspath(path)


double_exts = ('.tar.gz', '.nii.gz')


def split_extension(path):
    """
    A extension splitter that checks for compound extensions such as
    'file.nii.gz'

    Parameters
    ----------
    filename : str
        A filename to split into base and extension

    Returns
    -------
    base : str
        The base part of the string, i.e. 'file' of 'file.nii.gz'
    ext : str
        The extension part of the string, i.e. 'nii.gz' of 'file.nii.gz'
    """
    for double_ext in double_exts:
        if path.endswith(double_ext):
            return path[:-len(double_ext)], double_ext
    dirname = os.path.dirname(path)
    filename = os.path.basename(path)
    parts = filename.split('.')
    if len(parts) == 1:
        base = filename
        ext = None
    else:
        ext = '.' + parts[-1]
        base = '.'.join(parts[:-1])
    return os.path.join(dirname, base), ext


class classproperty(property):
    def __get__(self, cls, owner):
        return self.fget.__get__(None, owner)()


class NoContextWrapper(object):
    """
    Wraps an object, passing all calls through to the wrapped object
    except the __enter__ and __exit__ method, which do nothing. Used
    in cases where you want to use a file|connection handle within a
    "with" statement, except when it passed to the method from the
    calling code (presumably nested in another "with" statement).
    """

    def __init__(self, to_wrap):
        self._to_wrap = to_wrap

    def __getattr__(self, name):
        return getattr(self._to_wrap, name)

    def __enter__(self, *args, **kwargs):  # @UnusedVariable
        return self

    def __exit__(self, *args, **kwargs):
        pass



from .base import BaseRunner
from nipype.pipeline.plugins import MultiProcPlugin


class MultiProcRunner(BaseRunner):
    """
    Wrapper around the NiPype MultiProcPlugin

    Parameters
    ----------
    """

    nipype_plugin_cls = MultiProcPlugin

    def __init__(self, work_dir, num_processes=None, **kwargs):
        if num_processes is not None:
            kwargs['n_procs'] = num_processes
        super(MultiProcRunner, self).__init__(work_dir, **kwargs)

    @property
    def num_processes(self):
        return self._plugin.processes



import os.path
import subprocess as sp
import shutil
from unittest import TestCase
import errno
from xnat.exceptions import XNATError
import sys
import json
import filecmp
from collections import defaultdict
import warnings
import logging
import arcana
from arcana.utils import classproperty
from arcana.archive.local import (
    LocalArchive, SUMMARY_NAME)
from arcana.archive.xnat import download_all_datasets
from arcana.runner import LinearRunner
from arcana.exception import ArcanaError
from arcana.node import ArcanaNodeMixin
from arcana.exception import (
    ArcanaModulesNotInstalledException)
from traceback import format_exc
from arcana.archive.local import (
    SUMMARY_NAME as LOCAL_SUMMARY_NAME, FIELDS_FNAME)

logger = logging.getLogger('Arcana')
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
formatter = logging.Formatter("%(levelname)s - %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)

logging.getLogger("urllib3").setLevel(logging.WARNING)


class BaseTestCase(TestCase):

    SUBJECT = 'SUBJECT'
    VISIT = 'VISIT'
    SERVER = 'https://mbi-xnat.erc.monash.edu.au'
    XNAT_TEST_PROJECT = 'TEST001'
    REF_SUFFIX = '_REF'

    # The path to the test directory, which should sit along side the
    # the package directory. Note this will not work when Arcana
    # is installed by a package manager.
    BASE_TEST_DIR = os.path.abspath(os.path.join(
        os.path.dirname(arcana.__file__), '..', 'test'))

    @classproperty
    @classmethod
    def test_data_dir(cls):
        return os.path.join(cls.BASE_TEST_DIR, 'data')

    @classproperty
    @classmethod
    def unittest_root(cls):
        return os.path.join(cls.BASE_TEST_DIR, 'unittests')

    @classproperty
    @classmethod
    def archive_path(cls):
        return os.path.join(cls.test_data_dir, 'archive')

    @classproperty
    @classmethod
    def work_path(cls):
        return os.path.join(cls.test_data_dir, 'work')

    @classproperty
    @classmethod
    def base_cache_path(cls):
        return os.path.join(cls.test_data_dir, 'cache')

    def setUp(self, cache_dir=None):
        self.reset_dirs()
        self.add_session(self.project_dir, self.SUBJECT, self.VISIT,
                         cache_dir=cache_dir)

    def add_session(self, project_dir, subject, session,
                    required_datasets=None, cache_dir=None):
        if cache_dir is None:
            cache_dir = self.cache_dir
        session_dir = os.path.join(project_dir, subject, session)
        os.makedirs(session_dir)
        try:
            download_all_datasets(
                cache_dir, self.SERVER, self.xnat_session_name,
                overwrite=False)
        except Exception:
            if os.path.exists(cache_dir):
                warnings.warn(
                    "Could not download datasets from '{}_{}' session on "
                    "MBI-XNAT, attempting with what has already been "
                    "downloaded:\n\n{}"
                    .format(self.XNAT_TEST_PROJECT, self.name, format_exc()))
            else:
                raise
        for f in os.listdir(cache_dir):
            if required_datasets is None or f in required_datasets:
                src_path = os.path.join(cache_dir, f)
                dst_path = os.path.join(session_dir, f)
                if os.path.isdir(src_path):
                    shutil.copytree(src_path, dst_path)
                elif os.path.isfile(src_path):
                    shutil.copy(src_path, dst_path)
                else:
                    assert False
        # Download reference
        try:
            download_all_datasets(
                cache_dir, self.SERVER,
                self.xnat_session_name + self.REF_SUFFIX,
                overwrite=False)
        except Exception:
            pass

    def delete_project(self, project_dir):
        # Clean out any existing archive files
        shutil.rmtree(project_dir, ignore_errors=True)

    def reset_dirs(self):
        shutil.rmtree(self.project_dir, ignore_errors=True)
        shutil.rmtree(self.work_dir, ignore_errors=True)
        self.create_dirs()

    def create_dirs(self):
        for d in (self.project_dir, self.work_dir, self.cache_dir):
            if not os.path.exists(d):
                os.makedirs(d)

    @property
    def xnat_session_name(self):
        return '{}_{}'.format(self.XNAT_TEST_PROJECT, self.name)

    @property
    def session_dir(self):
        return self.get_session_dir(self.SUBJECT, self.VISIT)

    @property
    def cache_dir(self):
        return os.path.join(self.base_cache_path, self.name)

    @property
    def archive(self):
        return LocalArchive(self.project_dir)

    @property
    def runner(self):
        return LinearRunner(self.work_dir)

    @property
    def project_dir(self):
        return os.path.join(self.archive_path, self.name)

    @property
    def work_dir(self):
        return os.path.join(self.work_path, self.name)

    @property
    def name(self):
        return self._get_name(type(self))

    @property
    def project_id(self):
        return self.name  # To allow override in deriving classes

    def _get_name(self, cls):
        """
        Get unique name for test class from module path and its class name to
        be used for storing test data on XNAT and creating unique work/project
        dirs
        """
        module_path = os.path.abspath(sys.modules[cls.__module__].__file__)
        rel_module_path = module_path[(len(self.unittest_root) + 1):]
        path_parts = rel_module_path.split(os.path.sep)
        module_name = (''.join(path_parts[:-1]) +
                       os.path.splitext(path_parts[-1])[0][5:]).upper()
        test_class_name = cls.__name__[4:].upper()
        return module_name + '_' + test_class_name

    def create_study(self, study_cls, name, inputs, archive=None,
                     runner=None, **kwargs):
        """
        Creates a study using default archive and runners.

        Parameters
        ----------
        study_cls : Study
            The class to initialise
        name : str
            Name of the study
        inputs : List[BaseSpec]
            List of inputs to the study
        archive : BaseArchive | None
            The archive to use (a default local archive is used if one
            isn't provided
        runner : Runner | None
            The runner to use (a default LinearRunner is used if one
            isn't provided
        """
        if archive is None:
            archive = self.archive
        if runner is None:
            runner = self.runner
        return study_cls(
            name=name,
            archive=archive,
            runner=runner,
            inputs=inputs,
            **kwargs)

    def assertDatasetCreated(self, dataset_name, study_name, subject=None,
                             visit=None, frequency='per_session'):
        output_dir = self.get_session_dir(subject, visit, frequency)
        out_path = self.output_file_path(
            dataset_name, study_name, subject, visit, frequency)
        self.assertTrue(
            os.path.exists(out_path),
            ("Dataset '{}' (expected at '{}') was not created by unittest"
             " ('{}' found in '{}' instead)".format(
                 dataset_name, out_path, "', '".join(os.listdir(output_dir)),
                 output_dir)))

    def assertField(self, name, ref_value, study_name, subject=None,
                    visit=None, frequency='per_session',
                    to_places=None):
        esc_name = study_name + '_' + name
        output_dir = self.get_session_dir(subject, visit, frequency)
        try:
            with open(os.path.join(output_dir, FIELDS_FNAME)) as f:
                fields = json.load(f)
        except OSError as e:
            if e.errno == errno.ENOENT:
                raise ArcanaError(
                    "No fields were created by pipeline in study '{}'"
                    .format(study_name))
        try:
            value = fields[esc_name]
        except KeyError:
            raise ArcanaError(
                "Field '{}' was not created by pipeline in study '{}'. "
                "Created fields were ('{}')"
                .format(esc_name, study_name, "', '".join(fields)))
        msg = ("Field value '{}' for study '{}', {}, does not match "
               "reference value ({})".format(name, study_name, value,
                                             ref_value))
        if to_places is not None:
            self.assertAlmostEqual(
                value, ref_value, to_places,
                '{} to {} decimal places'.format(msg, to_places))
        else:
            self.assertEqual(value, ref_value, msg)

    def assertDatasetsEqual(self, dataset1, dataset2, error_msg=None):
        msg = "{} does not match {}".format(dataset1, dataset2)
        if msg is not None:
            msg += ':\n' + error_msg
        self.assertTrue(filecmp.cmp(dataset1.path, dataset2.path,
                                    shallow=False), msg=msg)

#     def assertImagesMatch(self, output, ref, study_name):
#         out_path = self.output_file_path(output, study_name)
#         ref_path = self.ref_file_path(ref)
#         try:
#             sp.check_output('diff {}.nii {}.nii'
#                             .format(out_path, ref_path), shell=True)
#         except sp.CalledProcessError as e:
#             if e.output == "Binary files {} and {} differ\n".format(
#                     out_path, ref_path):
#                 self.assert_(
#                     False,
#                     "Images {} and {} do not match exactly".format(out_path,
#                                                                    ref_path))
#             else:
#                 raise

    def assertStatEqual(self, stat, dataset_name, target, study_name,
                        subject=None, visit=None,
                        frequency='per_session'):
            try:
                ArcanaNodeMixin.load_module('mrtrix')
            except ArcanaModulesNotInstalledException:
                pass
            val = float(sp.check_output(
                'mrstats {} -output {}'.format(
                    self.output_file_path(
                        dataset_name, study_name,
                        subject=subject, visit=visit,
                        frequency=frequency),
                    stat),
                shell=True))
            self.assertEqual(
                val, target, (
                    "{} value of '{}' ({}) does not equal target ({}) "
                    "for subject {} visit {}"
                    .format(stat, dataset_name, val, target,
                            subject, visit)))

    def assertImagesAlmostMatch(self, out, ref, mean_threshold,
                                stdev_threshold, study_name):
        out_path = self.output_file_path(out, study_name)
        ref_path = self.ref_file_path(ref)
        # Should probably look into ITK fuzzy matching methods
        cmd = ("mrcalc -quiet {a} {b} -subtract - | mrstats - | "
               "grep -v channel | awk '{{print $4 \" \" $6}}'"
               .format(a=out_path, b=ref_path))
        out = sp.check_output(cmd, shell=True)
        mean, stdev = (float(x) for x in out.split())
        self.assert_(
            abs(mean) < mean_threshold and stdev < stdev_threshold,
            ("Mean ({mean}) or standard deviation ({stdev}) of difference "
             "between images {a} and {b} differ more than threshold(s) "
             "({thresh_mean} and {thresh_stdev} respectively)"
             .format(mean=mean, stdev=stdev, thresh_mean=mean_threshold,
                     thresh_stdev=stdev_threshold, a=out_path, b=ref_path)))

    def get_session_dir(self, subject=None, visit=None,
                        frequency='per_session'):
        if subject is None and frequency in ('per_session', 'per_subject'):
            subject = self.SUBJECT
        if visit is None and frequency in ('per_session', 'per_visit'):
            visit = self.VISIT
        if frequency == 'per_session':
            assert subject is not None
            assert visit is not None
            path = os.path.join(self.project_dir, subject, visit)
        elif frequency == 'per_subject':
            assert subject is not None
            assert visit is None
            path = os.path.join(
                self.project_dir, subject, SUMMARY_NAME)
        elif frequency == 'per_visit':
            assert visit is not None
            assert subject is None
            path = os.path.join(self.project_dir, SUMMARY_NAME, visit)
        elif frequency == 'per_project':
            assert subject is None
            assert visit is None
            path = os.path.join(self.project_dir, SUMMARY_NAME, SUMMARY_NAME)
        else:
            assert False
        return os.path.abspath(path)

    @classmethod
    def remove_generated_files(cls, study=None):
        # Remove derived datasets
        for fname in os.listdir(cls.get_session_dir()):
            if study is None or fname.startswith(study + '_'):
                os.remove(os.path.join(cls.get_session_dir(), fname))

    def output_file_path(self, fname, study_name, subject=None, visit=None,
                         frequency='per_session', **kwargs):
        return os.path.join(
            self.get_session_dir(subject=subject, visit=visit,
                                 frequency=frequency, **kwargs),
            '{}_{}'.format(study_name, fname))

    def ref_file_path(self, fname, subject=None, session=None):
        return os.path.join(self.session_dir, fname,
                            subject=subject, session=session)


class BaseMultiSubjectTestCase(BaseTestCase):

    SUMMARY_NAME = LOCAL_SUMMARY_NAME

    def setUp(self, cache_dir=None):
        self.reset_dirs()
        self.add_sessions(self.project_dir, cache_dir=cache_dir)

    def add_sessions(self, project_dir, required_datasets=None,
                     cache_dir=None):
        if cache_dir is None:
            cache_dir = self.cache_dir
        try:
            download_all_datasets(
                cache_dir, self.SERVER, self.xnat_session_name,
                overwrite=False)
        except XNATError as e:
            if os.path.exists(cache_dir):
                warnings.warn(
                    "Could not download datasets from '{}_{}' session on "
                    "MBI-XNAT, attempting with what has already been "
                    "downloaded:\n\n{}"
                    .format(self.XNAT_TEST_PROJECT, self.name, e))
            else:
                raise
        fnames = os.listdir(cache_dir)
        for fname in fnames:
            if fname == FIELDS_FNAME:
                continue
            if fname.startswith('.'):
                continue
            subj_id, visit_id, dataset = self._extract_ids(fname)
            if required_datasets is None or dataset in required_datasets:
                session_dir = self.make_session_dir(
                    project_dir, subj_id, visit_id)
                src_path = os.path.join(cache_dir, fname)
                dst_path = os.path.join(session_dir, dataset)
                if os.path.isdir(src_path):
                    shutil.copytree(src_path, dst_path)
                elif os.path.isfile(src_path):
                    shutil.copy(src_path, dst_path)
                else:
                    assert False
        if FIELDS_FNAME in fnames:
            fields = defaultdict(lambda: defaultdict(dict))
            with open(os.path.join(cache_dir, FIELDS_FNAME), 'rb') as f:
                all_fields = json.load(f)
            for name, value in all_fields.items():
                subj_id, visit_id, field_name = self._extract_ids(name)
                fields[subj_id][visit_id][field_name] = value
            for subj_id, subj_fields in fields.items():
                for visit_id, visit_fields in subj_fields.items():
                    session_dir = self.make_session_dir(
                        project_dir, subj_id, visit_id)
                    with open(os.path.join(session_dir, FIELDS_FNAME),
                              'w') as f:
                        json.dump(visit_fields, f)

    def _extract_ids(self, name):
        parts = name.split('_')
        if len(parts) < 3:
            raise ArcanaError(
                "'{}' in multi-subject test session '{}' needs to be "
                "prepended with subject and session IDs (delimited by "
                "'_')".format(name, self.xnat_session_name))
        subj_id, visit_id = parts[:2]
        if subj_id.lower() == SUMMARY_NAME.lower():
            subj_id = SUMMARY_NAME
        if visit_id.lower() == SUMMARY_NAME.lower():
            visit_id = SUMMARY_NAME
        basename = '_'.join(parts[2:])
        return subj_id, visit_id, basename

    @property
    def subject_ids(self):
        return (d for d in os.listdir(self.project_dir)
                if d != self.SUMMARY_NAME)

    def visit_ids(self, subject_id):
        subject_dir = os.path.join(self.project_dir, subject_id)
        return (d for d in os.listdir(subject_dir)
                if d != self.SUMMARY_NAME)

    def session_dir(self, subject, visit):
        return self.get_session_dir(subject, visit)

    def get_session_dir(self, subject, visit, frequency='per_session'):
        return super(BaseMultiSubjectTestCase, self).get_session_dir(
            subject=subject, visit=visit, frequency=frequency)

    @classmethod
    def make_session_dir(cls, project_dir, subj_id, visit_id):
        session_dir = os.path.join(project_dir, subj_id,
                                   visit_id)
        try:
            os.makedirs(session_dir)
        except OSError as e:
            if e.errno != errno.EEXIST:
                raise
        return session_dir


class DummyTestCase(BaseTestCase):

    def __init__(self):
        self.setUp()

    def __del__(self):
        self.tearDown()

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def assert_(self, statement, message=None):
        if not statement:
            message = "'{}' is not true".format(statement)
            print message
        else:
            print "Test successful"

    def assertEqual(self, first, second, message=None):
        if first != second:
            if message is None:
                message = '{} and {} are not equal'.format(repr(first),
                                                           repr(second))
            print message
        else:
            print "Test successful"

    def assertAlmostEqual(self, first, second, message=None):
        if first != second:
            if message is None:
                message = '{} and {} are not equal'.format(repr(first),
                                                           repr(second))
            print message
        else:
            print "Test successful"

    def assertLess(self, first, second, message=None):
        if first >= second:
            if message is None:
                message = '{} is not less than {}'.format(repr(first),
                                                          repr(second))
            print message
        else:
            print "Test successful"


class TestTestCase(BaseTestCase):

    def test_test(self):
        pass



from abc import ABCMeta, abstractmethod
from itertools import chain
from nipype.interfaces.base import (
    traits, DynamicTraitedSpec, Undefined, File, Directory,
    BaseInterface)
from arcana.node import Node
from arcana.dataset import (
    Dataset, DatasetSpec, FieldSpec, BaseField, BaseDataset)
from arcana.exception import ArcanaError
from arcana.utils import PATH_SUFFIX, FIELD_SUFFIX

PATH_TRAIT = traits.Either(File(exists=True), Directory(exists=True))
FIELD_TRAIT = traits.Either(traits.Int, traits.Float, traits.Str)
MULTIPLICITIES = ('per_session', 'per_subject', 'per_visit', 'per_project')


class Archive(object):
    """
    Abstract base class for all Archive systems, DaRIS, XNAT and local file
    system. Sets out the interface that all Archive classes should implement.
    """

    __metaclass__ = ABCMeta

    @abstractmethod
    def source(self, inputs, name=None, study_name=None, **kwargs):
        """
        Returns a NiPype node that gets the input data from the archive
        system. The input spec of the node's interface should inherit from
        ArchiveSourceInputSpec

        Parameters
        ----------
        project_id : str
            The ID of the project to return the sessions for
        inputs : list(Dataset|Field)
            An iterable of arcana.Dataset or arcana.Field
            objects, which specify the datasets to extract from the
            archive system
        name : str
            Name of the NiPype node
        study_name: str
            Prefix used to distinguish datasets generated by a particular
            study. Used for derived datasets only
        """
        if name is None:
            name = "{}_source".format(self.type)
        inputs = list(inputs)  # protected against iterators
        datasets = [i for i in inputs if isinstance(i, BaseDataset)]
        fields = [i for i in inputs if isinstance(i, BaseField)]
        return Node(self.Source(study_name, datasets, fields, **kwargs),
                    name=name)

    @abstractmethod
    def sink(self, outputs, frequency='per_session', name=None,
             study_name=None, **kwargs):
        """
        Returns a NiPype node that puts the output data back to the archive
        system. The input spec of the node's interface should inherit from
        ArchiveSinkInputSpec

        Parameters
        ----------
        project_id : str
            The ID of the project to return the sessions for
        outputs : List(BaseFile|Field) | list(
            An iterable of arcana.Dataset arcana.Field objects,
            which specify the datasets to put into the archive system
        name : str
            Name of the NiPype node
        study_name: str
            Prefix used to distinguish datasets generated by a particular
            study. Used for derived datasets only

        """
        if name is None:
            name = "{}_{}_sink".format(self.type, frequency)
        outputs = list(outputs)  # protected against iterators
        if frequency.startswith('per_session'):
            sink_class = self.Sink
        elif frequency.startswith('per_subject'):
            sink_class = self.SubjectSink
        elif frequency.startswith('per_visit'):
            sink_class = self.VisitSink
        elif frequency.startswith('per_project'):
            sink_class = self.ProjectSink
        else:
            raise ArcanaError(
                "Unrecognised frequency '{}' can be one of '{}'"
                .format(frequency,
                        "', '".join(Dataset.MULTIPLICITY_OPTIONS)))
        datasets = [o for o in outputs if isinstance(o, BaseDataset)]
        fields = [o for o in outputs if isinstance(o, BaseField)]
        return Node(sink_class(study_name, datasets, fields, **kwargs),
                    name=name)

    def __ne__(self, other):
        return not (self == other)


class BaseArchiveNode(BaseInterface):
    """
    Parameters
    ----------
    infields : list of str
        Indicates the input fields to be dynamically created

    outfields: list of str
        Indicates output fields to be dynamically created

    See class examples for usage

    """

    def __init__(self, study_name, datasets, fields):
        super(BaseArchiveNode, self).__init__()
        self._study_name = study_name
        self._datasets = datasets
        self._fields = fields

    def __eq__(self, other):
        try:
            return (self.study_name == other.study_name and
                    self.datasets == other.datasets and
                    self.fields == other.fields)
        except AttributeError:
            return False

    def __repr__(self):
        return "{}(study_name='{}', datasets={}, fields={})".format(
            type(self).__name__, self.study_name, self.datasets,
            self.fields)

    def __ne__(self, other):
        return not self == other

    def _run_interface(self, runtime, *args, **kwargs):  # @UnusedVariable
        return runtime

    @property
    def study_name(self):
        return self._study_name

    @property
    def datasets(self):
        return self._datasets

    @property
    def fields(self):
        return self._fields

    @classmethod
    def _add_trait(cls, spec, name, trait_type):
        spec.add_trait(name, trait_type)
        spec.trait_set(trait_change_notify=False, **{name: Undefined})
        # Access the trait (not sure why but this is done in add_traits
        # so I have also done it here
        getattr(spec, name)

    def prefix_study_name(self, name, is_spec=True):
        """Prepend study name if defined"""
        if is_spec:
            name = self.study_name + '_' + name
        return name


class ArchiveSourceInputSpec(DynamicTraitedSpec):
    """
    Base class for archive source input specifications. Provides a common
    interface for 'run_pipeline' when using the archive source to extract
    primary and preprocessed datasets from the archive system
    """
    subject_id = traits.Str(mandatory=True, desc="The subject ID")
    visit_id = traits.Str(mandatory=True, usedefult=True,
                            desc="The visit or derived group ID")


class ArchiveSource(BaseArchiveNode):
    """
    Parameters
    ----------
    datasets: list
        List of all datasets to be extracted from the archive
    fields: list
        List of all the fields that are to be extracted from the archive
    study_name: str
        Prefix prepended onto derived dataset "names"
    """

    output_spec = DynamicTraitedSpec
    _always_run = True

    def _outputs(self):
        outputs = super(ArchiveSource, self)._outputs()
        # Add output datasets
        for dataset in self.datasets:
            assert isinstance(dataset, BaseDataset)
            self._add_trait(outputs, dataset.name + PATH_SUFFIX,
                            PATH_TRAIT)
        # Add output fields
        for field in self.fields:
            assert isinstance(field, BaseField)
            self._add_trait(outputs, field.name + FIELD_SUFFIX,
                            field.dtype)
        return outputs


class BaseArchiveSinkSpec(DynamicTraitedSpec):
    pass


class ArchiveSinkInputSpec(BaseArchiveSinkSpec):

    subject_id = traits.Str(mandatory=True, desc="The subject ID"),
    visit_id = traits.Str(mandatory=False,
                            desc="The session or derived group ID")


class ArchiveSubjectSinkInputSpec(BaseArchiveSinkSpec):

    subject_id = traits.Str(mandatory=True, desc="The subject ID")


class ArchiveVisitSinkInputSpec(BaseArchiveSinkSpec):

    visit_id = traits.Str(mandatory=True, desc="The visit ID")


class ArchiveProjectSinkInputSpec(BaseArchiveSinkSpec):
    pass


class BaseArchiveSinkOutputSpec(DynamicTraitedSpec):

    out_files = traits.List(PATH_TRAIT, desc='Output datasets')

    out_fields = traits.List(
        traits.Tuple(traits.Str, FIELD_TRAIT), desc='Output fields')


class ArchiveSinkOutputSpec(BaseArchiveSinkOutputSpec):

    subject_id = traits.Str(desc="The subject ID")
    visit_id = traits.Str(desc="The visit ID")


class ArchiveSubjectSinkOutputSpec(BaseArchiveSinkOutputSpec):

    subject_id = traits.Str(desc="The subject ID")


class ArchiveVisitSinkOutputSpec(BaseArchiveSinkOutputSpec):

    visit_id = traits.Str(desc="The visit ID")


class ArchiveProjectSinkOutputSpec(BaseArchiveSinkOutputSpec):

    project_id = traits.Str(desc="The project ID")


class BaseArchiveSink(BaseArchiveNode):

    def __init__(self, study_name, datasets, fields):
        super(BaseArchiveSink, self).__init__(study_name, datasets,
                                              fields)
        # Add input datasets
        for dataset in datasets:
            assert isinstance(dataset, DatasetSpec)
            self._add_trait(self.inputs, dataset.name + PATH_SUFFIX,
                            PATH_TRAIT)
        # Add input fields
        for field in fields:
            assert isinstance(field, FieldSpec)
            self._add_trait(self.inputs, field.name + FIELD_SUFFIX,
                            field.dtype)


class ArchiveSink(BaseArchiveSink):

    input_spec = ArchiveSinkInputSpec
    output_spec = ArchiveSinkOutputSpec

    frequency = 'per_session'

    def _base_outputs(self):
        outputs = self.output_spec().get()
        outputs['subject_id'] = self.inputs.subject_id
        outputs['visit_id'] = self.inputs.visit_id
        return outputs


class ArchiveSubjectSink(BaseArchiveSink):

    input_spec = ArchiveSubjectSinkInputSpec
    output_spec = ArchiveSubjectSinkOutputSpec

    frequency = 'per_subject'

    def _base_outputs(self):
        outputs = self.output_spec().get()
        outputs['subject_id'] = self.inputs.subject_id
        return outputs


class ArchiveVisitSink(BaseArchiveSink):

    input_spec = ArchiveVisitSinkInputSpec
    output_spec = ArchiveVisitSinkOutputSpec

    frequency = 'per_visit'

    def _base_outputs(self):
        outputs = self.output_spec().get()
        outputs['visit_id'] = self.inputs.visit_id
        return outputs


class ArchiveProjectSink(BaseArchiveSink):

    input_spec = ArchiveProjectSinkInputSpec
    output_spec = ArchiveProjectSinkOutputSpec

    frequency = 'per_project'

    def _base_outputs(self):
        outputs = self.output_spec().get()
        return outputs


class Project(object):

    def __init__(self, subjects, visits, datasets, fields):
        self._subjects = subjects
        self._visits = visits
        self._datasets = datasets
        self._fields = fields

    @property
    def subjects(self):
        return iter(self._subjects)

    @property
    def visits(self):
        return iter(self._visits)

    @property
    def datasets(self):
        return self._datasets

    @property
    def fields(self):
        return self._fields

    @property
    def dataset_names(self):
        return (d.name for d in self.datasets)

    @property
    def field_names(self):
        return (f.name for f in self.fields)

    @property
    def data(self):
        return chain(self.datasets, self.fields)

    def nodes(self, frequency):
        if frequency == 'per_session':
            nodes = chain(*(s.sessions for s in self.subjects))
        elif frequency == 'per_subject':
            nodes = self.subjects
        elif frequency == 'per_visit':
            nodes = self.visits
        elif frequency == 'per_project':
            nodes = [self]
        else:
            assert False
        return nodes

    @property
    def data_names(self):
        return (d.name for d in self.data)

    def __eq__(self, other):
        if not isinstance(other, Project):
            return False
        return (self._subjects == other._subjects and
                self._visits == other._visits and
                self._datasets == other._datasets and
                self._fields == other._fields)

    def find_mismatch(self, other, indent=''):
        """
        Used in debugging unittests
        """
        if self != other:
            mismatch = "\n{}Project".format(indent)
        else:
            mismatch = ''
        sub_indent = indent + '  '
        if len(list(self.subjects)) != len(list(other.subjects)):
            mismatch += ('\n{indent}mismatching subject lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.subjects)),
                                 len(list(other.subjects)),
                                 list(self.subjects),
                                 list(other.subjects),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.subjects, other.subjects):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        if len(list(self.visits)) != len(list(other.visits)):
            mismatch += ('\n{indent}mismatching visit lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.visits)),
                                 len(list(other.visits)),
                                 list(self.visits),
                                 list(other.visits),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.visits, other.visits):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        if len(list(self.datasets)) != len(list(other.datasets)):
            mismatch += ('\n{indent}mismatching summary dataset lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.datasets)),
                                 len(list(other.datasets)),
                                 list(self.datasets),
                                 list(other.datasets),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.datasets, other.datasets):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        if len(list(self.fields)) != len(list(other.fields)):
            mismatch += ('\n{indent}mismatching summary field lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.fields)),
                                 len(list(other.fields)),
                                 list(self.fields),
                                 list(other.fields),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.fields, other.fields):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        return mismatch

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return ("Project(num_subjects={}, num_visits={}, "
                "num_datasets={}, num_fields={})".format(
                    len(list(self.subjects)),
                    len(list(self.visits)),
                    len(list(self.datasets)), len(list(self.fields))))


class Subject(object):
    """
    Holds a subject id and a list of sessions
    """

    def __init__(self, subject_id, sessions, datasets, fields):
        self._id = subject_id
        self._sessions = sessions
        self._datasets = datasets
        self._fields = fields
        for session in sessions:
            session.subject = self

    @property
    def id(self):
        return self._id

    def __lt__(self, other):
        return self._id < other._id

    @property
    def sessions(self):
        return iter(self._sessions)

    @property
    def datasets(self):
        return self._datasets

    @property
    def fields(self):
        return self._fields

    @property
    def dataset_names(self):
        return (d.name for d in self.datasets)

    @property
    def field_names(self):
        return (f.name for f in self.fields)

    @property
    def data(self):
        return chain(self.datasets, self.fields)

    @property
    def data_names(self):
        return (d.name for d in self.data)

    def __eq__(self, other):
        if not isinstance(other, Subject):
            return False
        return (self._id == other._id and
                self._sessions == other._sessions and
                self._datasets == other._datasets and
                self._fields == other._fields)

    def find_mismatch(self, other, indent=''):
        if self != other:
            mismatch = "\n{}Subject '{}' != '{}'".format(
                indent, self.id, other.id)
        else:
            mismatch = ''
        sub_indent = indent + '  '
        if self.id != other.id:
            mismatch += ('\n{}id: self={} v other={}'
                         .format(sub_indent, self.id, other.id))
        if len(list(self.sessions)) != len(list(other.sessions)):
            mismatch += ('\n{indent}mismatching session lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.sessions)),
                                 len(list(other.sessions)),
                                 list(self.sessions),
                                 list(other.sessions),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.sessions, other.sessions):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        if len(list(self.datasets)) != len(list(other.datasets)):
            mismatch += ('\n{indent}mismatching summary dataset lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.datasets)),
                                 len(list(other.datasets)),
                                 list(self.datasets),
                                 list(other.datasets),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.datasets, other.datasets):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        if len(list(self.fields)) != len(list(other.fields)):
            mismatch += ('\n{indent}mismatching summary field lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.fields)),
                                 len(list(other.fields)),
                                 list(self.fields),
                                 list(other.fields),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.fields, other.fields):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        return mismatch

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return ("Subject(id={}, num_sessions={})"
                .format(self._id, len(self._sessions)))


class Visit(object):
    """
    Holds a subject id and a list of sessions
    """

    def __init__(self, visit_id, sessions, datasets, fields):
        self._id = visit_id
        self._sessions = sessions
        self._datasets = datasets
        self._fields = fields
        for session in sessions:
            session.visit = self

    @property
    def id(self):
        return self._id

    def __lt__(self, other):
        return self._id < other._id

    @property
    def sessions(self):
        return iter(self._sessions)

    @property
    def datasets(self):
        return self._datasets

    @property
    def fields(self):
        return self._fields

    @property
    def dataset_names(self):
        return (d.name for d in self.datasets)

    @property
    def field_names(self):
        return (f.name for f in self.fields)

    @property
    def data(self):
        return chain(self.datasets, self.fields)

    @property
    def data_names(self):
        return (d.name for d in self.data)

    def __eq__(self, other):
        if not isinstance(other, Visit):
            return False
        return (self._id == other._id and
                self._sessions == other._sessions and
                self._datasets == other._datasets and
                self._fields == other._fields)

    def find_mismatch(self, other, indent=''):
        if self != other:
            mismatch = "\n{}Visit '{}' != '{}'".format(
                indent, self.id, other.id)
        else:
            mismatch = ''
        sub_indent = indent + '  '
        if self.id != other.id:
            mismatch += ('\n{}id: self={} v other={}'
                         .format(sub_indent, self.id, other.id))
        if len(list(self.sessions)) != len(list(other.sessions)):
            mismatch += ('\n{indent}mismatching session lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.sessions)),
                                 len(list(other.sessions)),
                                 list(self.sessions),
                                 list(other.sessions),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.sessions, other.sessions):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        if len(list(self.datasets)) != len(list(other.datasets)):
            mismatch += ('\n{indent}mismatching summary dataset lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.datasets)),
                                 len(list(other.datasets)),
                                 list(self.datasets),
                                 list(other.datasets),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.datasets, other.datasets):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        if len(list(self.fields)) != len(list(other.fields)):
            mismatch += ('\n{indent}mismatching summary field lengths '
                         '(self={} vs other={}): '
                         '\n{indent}    self={}\n{indent}    other={}'
                         .format(len(list(self.fields)),
                                 len(list(other.fields)),
                                 list(self.fields),
                                 list(other.fields),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.fields, other.fields):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        return mismatch

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return "Visit(id={}, num_sessions={})".format(self._id,
                                                      len(self._sessions))


class Session(object):
    """
    Holds the session id and the list of datasets loaded from it

    Parameters
    ----------
    subject_id : str
        The subject ID of the session
    visit_id : str
        The visit ID of the session
    datasets : list(Dataset)
        The datasets found in the session
    derived : Session
        If derived scans are stored in a separate session, it is provided
        here
    """

    def __init__(self, subject_id, visit_id, datasets, fields, derived=None):
        self._subject_id = subject_id
        self._visit_id = visit_id
        self._datasets = datasets
        self._fields = fields
        self._subject = None
        self._visit = None
        self._derived = derived

    @property
    def visit_id(self):
        return self._visit_id

    @property
    def subject_id(self):
        return self._subject_id

    def __lt__(self, other):
        if self.subject_id < other.subject_id:
            return True
        else:
            return self.visit_id < other.visit_id

    @property
    def subject(self):
        return self._subject

    @subject.setter
    def subject(self, subject):
        self._subject = subject

    @property
    def visit(self):
        return self._visit

    @visit.setter
    def visit(self, visit):
        self._visit = visit

    @property
    def derived(self):
        return self._derived

    @derived.setter
    def derived(self, derived):
        self._derived = derived

    @property
    def acquired(self):
        """True if the session contains acquired scans"""
        return not self._derived or self._derived is None

    @property
    def datasets(self):
        return self._datasets

    @property
    def fields(self):
        return self._fields

    @property
    def dataset_names(self):
        return (d.name for d in self.datasets)

    @property
    def field_names(self):
        return (f.name for f in self.fields)

    @property
    def data(self):
        return chain(self.datasets, self.fields)

    @property
    def data_names(self):
        return (d.name for d in self.data)

    @property
    def derived_dataset_names(self):
        datasets = (self.datasets
                    if self.derived is None else self.derived.datasets)
        return (d.name for d in datasets)

    @property
    def derived_field_names(self):
        fields = (self.fields
                  if self.derived is None else self.derived.fields)
        return (f.name for f in fields)

    @property
    def derived_data_names(self):
        return chain(self.derived_dataset_names,
                     self.derived_field_names)

    @property
    def all_dataset_names(self):
        return chain(self.dataset_names, self.derived_dataset_names)

    @property
    def all_field_names(self):
        return chain(self.field_names, self.derived_field_names)

    @property
    def all_data_names(self):
        return chain(self.data_names, self.derived_data_names)

    def __eq__(self, other):
        if not isinstance(other, Session):
            return False
        return (self.subject_id == other.subject_id and
                self.visit_id == other.visit_id and
                self.datasets == other.datasets and
                self.fields == other.fields and
                self.derived == other.derived)

    def find_mismatch(self, other, indent=''):
        if self != other:
            mismatch = "\n{}Session '{}-{}' != '{}-{}'".format(
                indent, self.subject_id, self.visit_id,
                other.subject_id, other.visit_id)
        else:
            mismatch = ''
        sub_indent = indent + '  '
        if self.subject_id != other.subject_id:
            mismatch += ('\n{}subject_id: self={} v other={}'
                         .format(sub_indent, self.subject_id,
                                 other.subject_id))
        if self.visit_id != other.visit_id:
            mismatch += ('\n{}visit_id: self={} v other={}'
                         .format(sub_indent, self.visit_id,
                                 other.visit_id))
        if self.derived != other.derived:
            mismatch += ('\n{}derived: self={} v other={}'
                         .format(sub_indent, self.derived,
                                 other.derived))
        if len(list(self.datasets)) != len(list(other.datasets)):
            mismatch += ('\n{indent}mismatching dataset lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.datasets)),
                                 len(list(other.datasets)),
                                 list(self.datasets),
                                 list(other.datasets),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.datasets, other.datasets):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        if len(list(self.fields)) != len(list(other.fields)):
            mismatch += ('\n{indent}mismatching field lengths '
                         '(self={} vs other={}): '
                         '\n{indent}  self={}\n{indent}  other={}'
                         .format(len(list(self.fields)),
                                 len(list(other.fields)),
                                 list(self.fields),
                                 list(other.fields),
                                 indent=sub_indent))
        else:
            for s, o in zip(self.fields, other.fields):
                mismatch += s.find_mismatch(o, indent=sub_indent)
        return mismatch

    def __ne__(self, other):
        return not (self == other)

    def __repr__(self):
        return ("Session(subject_id='{}', visit_id='{}', num_datasets={}, "
                "num_fields={}, derived={})".format(
                    self.subject_id, self.visit_id, len(self._datasets),
                    len(self._fields), self.derived))
